#include <iostream>
#include <fstream>
#include <vector>
#include <stack> 
#include <map>
#include <algorithm>
#include <unistd.h>
#include "string.h"
#include "ppn.h"
using namespace std;

int debug_level = 0;                       // уровень отладочного вывода
//-------------------------- описание типов данных в языке --------------------------
enum type_t { UNDEF = -1, NUMBER = 0, STRING };
struct type {                              // описание типов переменных в языке
   string title;
   type_t code;
   string nil;                             // нулевое значение этого типа данных
   bool (*test)( const string& );          // тестирование формата записи константы
   string (*cvalue)( const string& );      // значение из записи константы в коде
} const types[] = { 
                     { "int",    NUMBER, "0", test_int,    
                       []( const string& s )-> string { return s; } }, 
                     { "string", STRING, "",  test_string,
                       []( const string& s )-> string { return s.substr( 1, s.length() - 2 ); } } 
                  };
const string& type_string( const type_t type ) {
   static const string undef( "undef" ); 
   for( auto &s : types )
      if( type == s.code ) return s.title;      
   return undef;      
} 
//----------- определения операций и их действий для всех типов ---------------------
#include "ops.h"
//----------------- описание значения (переменной или константы) --------------------
struct variable {                          // значение (переменной или константы)
   type_t type;                            // код типа данных
   string value;
public:
   variable( type_t type, string value ) : type( type ), value( value ) {}
};
//----------------------------- класс кода программы --------------------------------
class progz {
private:
   vector<string> input;                   // список входных строк программы
   map<string,variable*> var_table;        //**************  таблица описанных переменных
   string inline err_msg( const string& msg ) {
      return "синтаксис - " + msg;
   }
   variable* find_var( const string& s ) { // поиск переменной в таблице
      try { return var_table.at( s ); } 
      catch(...) { return NULL; }          // исключение если не найдена переменная
   }
   map<string,unsigned> lbl_table;         //**************  таблица меток строк 
   int find_lbl( const string& s ) {       // поиск позиции помеченного оператора
      try { return lbl_table.at( s ); } 
      catch(...) { return -1; }            // исключение если не найдена метка
   }
   bool label( string& l ) {
      int lp = l.find( ":" );
      if( lp > 0 ) {                       // помеченная строка оператора?
         string title = strtrim( l.substr( 0, lp ) );
         l = strtrim( l.substr( lp + 1 ) );
         lbl_table.insert( pair< string, unsigned >( title, opos ) );
      } 
      return lp > 0;
   }
   void add_vars( type_t t, string l ) {     // разбор строки описаний переменных
      vector<string> ops = strsep( l, "," );
      for( string &op : ops ) {
         if( label( op ) )
            throw syn_exception( err_msg( "метка в описании переменных" ) );           
         vector<string> var = strsep( op, "=" );
         if( var.size() > 2 )
            throw syn_exception( err_msg( "цепочка инициализаций" ) );           
         if( !test_var( var[ 0 ] ) )
            throw syn_exception( err_msg( "недопустимое имя переменной: " + var[ 0 ] ) ); 
         if( find_var( var[ 0 ] ) )
            throw syn_exception( err_msg( "повторное определение переменной: " + var[ 0 ] ) ); 
         string value( types[ t ].nil );   // пустое значение типа
         if( var.size() > 1 ) {            // инициализация
             if( !types[ t ].test( var[ 1 ] ) )
                throw syn_exception( err_msg( "не соответствует тип константы: " + 
                                     type_string( t ) + " " +  var[ 1 ] ) ); 
             value = types[ t ].cvalue( var[ 1 ] ); // значение константы
         }; 
         var_table.insert( pair< string, variable* >( var[ 0 ], new variable( t, value ) ) );
         if( debug_level > 1 ) {
            variable *v = var_table[ var[ 0 ] ];
            cout << "+ переменная " << var[ 0 ] << " = " << type_string( v->type ) << "  " << v->value << endl;  
         }
      }    
   }
   vector<vector<string>> output;          //************** внутреннее представление кода
   int opos;                               // номер позиции перехода (для GOTO)
   static const string cmd[];
   static const int cmd_size;
   enum cmd_t { read, write, go_to };
   void add_opers( string l ) {            // разбор операторных строк
      vector<string> eval;                 // запись оператора по токенам 
      label( l );                          // есть метка? ... поместить
      const string rd( cmd[ read ] ), wr( cmd[ write ] ), gt( cmd[ go_to ] );
      if( !l.find( rd + "(" ) ) {
         if( ')' != *( l.end() - 1 ) )
            throw syn_exception( err_msg( "закрывающаяся скобка read" ) );
         l = strtrim( l.substr( rd.length() + 1, l.length() - rd.length() - 2 ) );
         if( !test_var( l ) )
            throw syn_exception( err_msg( "не L-значение в read: " + l ) ); 
         if( !find_var( l ) )
            throw syn_exception( err_msg( "не определённая переменная: " + l ) ); 
         eval.push_back( rd );             // <read z> - ввод переменной z
         eval.push_back( l );
         output.push_back( eval );
         opos++;
         return;
      } 
      if( !l.find( gt + " " ) ) {
         l = strtrim( l.substr( gt.length() + 1 ) );
         eval.push_back( gt );             // <goto z> - ввод переменной z
         eval.push_back( l );
         output.push_back( eval );
         opos++;
         return;
      }
      str2ppn pp; 
      if( !l.find( wr + "(" ) ) {
         if( ')' != *( l.end() - 1 ) )
            throw syn_exception( err_msg( "закрывающаяся скобка write" ) );
         l =  strtrim( l.substr( wr.length() + 1, l.length() - wr.length() - 2 ) );
         vector<string> ops = strsep( l, "," );
         if( debug_level > 1 )cout << "цепочка write длиной: " << ops.size() << endl;
         try {
            for( string &op : ops ) {
               eval.clear();
               eval = pp.convert( op );      // <xxx write> - вывод вычисленного значения xxx
               eval.push_back( wr ); 
               output.push_back( eval );
               opos++;
            }
         }
         catch(...) { throw; }
         return;
      }
      vector<string> var = strsep( l, "=" ); // обработка присвоений
      int n = var.size();
      if( n-- < 2 ) return;
      try { 
         eval = pp.convert( var[ n ] );      // вычисление последнего выражения в цепочке
      }
      catch(...) { throw; }
      while( n-- > 0 ) {
         if( !find_var( var[ n ] ) )
            throw syn_exception( err_msg( "не определеная переменная: " + var[ 0 ] ) ); 
         eval.push_back( "=" );
         eval.push_back( var[ n ] );
      }
      output.push_back( eval );
      opos++;
   }
   void prepare( void ) {                  // лексический разбор исходного кода
      var_table.clear();
      lbl_table.clear();
      opos = 0;                            // начальная позиция внутренного представления
      for( string &l : input ) {
         type_t type = UNDEF;              // операторная строка (не описания переменных)
         for( auto &ct : types )           // ищем заголовочный описатель типа переменных
            if( !l.find( ct.title + " " ) ) {
               type = ct.code;
               l = strtrim( l.substr( ct.title.length() ) );
               break;
            } 
         if( debug_level > 1 )
            cout << ( type < 0 ? '-' : '+' ) << "\t| " << l << endl;
         try {
            if( type < 0 ) add_opers( l ); // разбор строки описаний (переменных) 
            else add_vars( type, l );      // разбор строки операторов 
         }
         catch(...) { throw; }
      }
      begin();                             // установить указатель на начало 
   }
   void put_step( vector<vector<string>>::iterator is ) {
      for( auto &o : *is ) 
         cout << o << ' ';
      cout << endl;   
   }   
   void eval( const vector<string>& ops ) {            // выполнение одного оператора-строки
      stack<string> stak;                              //************** вычислительный стек
      vector<string>::const_iterator p = ops.begin();
      opos = -1;                                       // последовательное выполнение, без GOTO
      type_t type = UNDEF;                             // тип операндов в этом выражении
      while( p <= ops.end() ) {
         if( debug_level > 1 )
            cout << *p << "[" << p->length() << "] глубина стека = " << stak.size() << endl;
         type_t t;                                     // переменная для выяснения типа объктов выражения
         string res;
         variable* v; 
         for( int i = 0; i < cmd_size; i++ ) {         // проверка на команды...
            if( *p == cmd[ write ] ) {                 // команда write()  
               cout << stak.top() << endl; 
               stak.pop();
               return;               
            }
            if( *p == cmd[ read ] ) {                  // команда read()
               cout << "> ";
               getline( cin, res );
               if( !( v = find_var( *++p ) ) )         // запись будет в следующий токен
                  throw syn_exception( run_msg( "неизвестная переменная: " + *p ) );
               t = v->type;
               if( types[ t ].test( res ) )            // проверка соответствию типа
                  v->value = res;
               else 
                  throw syn_exception( run_msg( "несоответствие типа введенного" ) );
               return;
            }
            if( *p == cmd[ go_to ] ) {                 // команда goto   
               opos = find_lbl( *++p );
               if( opos < 0 )
                  throw syn_exception( run_msg( "не найдена метка: " + *p ) );
               return;               
            }            
         }
         if( "=" == *p ) {                             // операция =   
            if( !( v = find_var( *++p ) ) )
               throw syn_exception( run_msg( "неизвестная переменная: " + *p ) );
            v->value = stak.top();
            if( ++p == ops.end() ) {                   // последнее присвоение в цепочке  
               stak.pop();
               return; 
            }
            if( "=" == *p )                            // цепочечное присвоение
               continue;
            throw syn_exception( run_msg( "неверная цепочка присвоений" ) );
         }
         op_func eval = opers::find( *p );               // найден знак операции в таблице
         if( eval ) {                                  
               string op1, op2;
               if( stak.size() < 2 )
                  throw syn_exception( run_msg( "незаполненный стек" ) );
               op2 = stak.top();
               stak.pop();
               op1 = stak.top();
               stak.pop();
               try {
                  res = eval( type, op1, op2 );        // выполнение операции
               }
               catch( syn_exception& e ) {             // ошибки выполнения операции      
                  cout << run_msg( (const char*)e ) << endl;  
                  return;
               } 
               if( debug_level > 1 )
                  cout << "результат вычисления в стек <= " << res << endl;
               stak.push( res );
               p++;
               continue;
         }
         string val = *p;                              // операнд
         bool cnst = false;
         for( auto &ct : types ) 
            if( ( cnst = ct.test( val ) ) ) {          // это похоже на константу какого типа?   
               t = ct.code;
               break;
            }
         if( cnst && debug_level > 2 ) 
            cout << "найдена константа " << type_string( t ) << " с значением " << val << endl;
         if( !cnst ) {
            if( !( v = find_var( *p ) ) )              // переменная?
               throw syn_exception( run_msg( "неизвестная переменная: " + *p ) );
            if( debug_level > 2 ) 
               cout << "найдена переменная " << *p << " тип " << type_string( v->type ) 
                    << " с значением " << v->value << endl;  
            t = v->type;
            val = v->value;
         }
         if( UNDEF == type ) {                         // выяснение типа операнда (неопределённый) 
            type = t;
            if( debug_level > 2 )
               cout << "устанавливается тип выражения = " << type_string( t ) << endl;
         }
         else {                                        // проверка соответствия типа
            if( debug_level > 2 )
               cout << "проверяется тип операнда = " << type_string( t ) << endl;
            if( t != type )
               throw syn_exception( run_msg( "смешение типов операндов" ) );
         }
         if( debug_level > 1 )
            cout << "операнд заносится в стек <= " << val << endl;
         stak.push( val );
         p++;
      }
   }
   vector<vector<string>>::iterator ip;    // указатель (итератор) текущей инструкции
public:
   bool do_step( bool step = true ) {      // один шаг выполнения
      if( ip < output.end() ) {
         if( debug_level && step ) {  
            cout << "[" << ip - output.begin() << "]: "; 
            put_step( ip );
         }
      }
      else {                               // программа выполнена!
         if( step )                        // сообщать только в пошаговом режиме
            cout << run_msg( "достигнут конец программы" ) << endl; 
         return false;
      }
      try {                                // ошибки выполнения
         eval( *ip );                      // только при успехе
         if( opos < 0 )                    // не GOTO
            ip++;
         else 
            ip = output.begin() + opos; 
      }
      catch( syn_exception& e ) {         
         string s = string( (const char*)e );
         cout << s << endl;
      }
      return true;
   }
   void exec( int n = 10 ) {               // выполнить от начала до конца
      begin();
      while( do_step( false ) ) {
         if( n > 0 )
            if( 0 == --n ) break;
      }
   }
   void clean( void ) {                    // удаление описаний переменных
      for( auto t : var_table ) 
         delete t.second;
      var_table.clear();
   }
   string inline run_msg( const string& msg ) {
      return "выполнение - " + msg;
   }
   void inline begin( void ) {             // начать выполнение с начала
      ip = output.begin(); 
   }
   void put_vars( void ) {                 // индикация таблицы переменных
      for( auto &v : var_table )
         cout << v.first << '\t'
              << type_string( v.second-> type ) << '\t'
              << '<' << v.second->value << '>' << endl;
   }
   void put_opers( void ) {                // индикация внутреннего представления кода
      for( auto is = output.begin(); is != output.end(); is++ )
         put_step( is );
   }
   void put_lbls( void ) {                 // индикация меток операторов
      for( auto &v : lbl_table )
         cout << v.first << '\t' << v.second << endl;
   }
   friend inline ostream& operator <<( ostream& out, progz& z ) {
      return out << z.input;
   }
   inline friend istream& operator >>( istream& inp, progz& z ) {
      if( debug_level > 1 ) cout << "--- ввод старт ---" << endl;
      string text;
      z.input.clear();
      z.clean();
      while( true ) {                      // ввод текста программы -> z.text
         string cur;
         getline( inp, cur );
         if( !inp ) break;                 // EOF
         int cmnt;
         if( ( cmnt = cur.find( "//" ) ) >= 0 ) {
            cur = strtrim( cur.substr( 0, cmnt ) );
         }
         if( !cur.length() ) continue;     // пустая строка
         text += strtrim( cur ); 
      } 
      const string sprogram = "program";
      int pos = text.find( sprogram );
      if( pos ) {                          // не начальная позиция
         cout << z.run_msg( "не начальная позиция " + sprogram ) << endl; 
         return inp;
      }
      text = text.substr( sprogram.length() );
      text = strtrim( text );
      if( text[ 0 ] != '{' || text[ text.length() - 1 ] != '}' ) {
         cout << z.run_msg( "баланс скобок " + sprogram + " {...}" ) << endl; 
         return inp;
      }
      text = text.substr( 1, text.length() - 2 );
      text = strtrim( text );
      z.input = strsep( text, ";" );
      z.output.clear();
      try {
         z.prepare();                      // лексический разбор 
      }
      catch( syn_exception& e ) {          // синтаксическая ошибка
         string s = string( (const char*)e );
         cout << s << endl;
         z.input.clear();
         z.output.clear();
      }
      if( debug_level > 1 ) cout << "--- ввод завершение ---" << endl;
      if( debug_level > 2 ) cout << "размер внутреннего кода: " << z.output.size() << endl;
      return inp;
   }
};
const string progz::cmd[] = {              // текст команд 
   "read", "write", "goto" 
};
const int progz::cmd_size = sizeof( cmd ) / sizeof( cmd[ 0 ] );
//------------------------------------------------------------------------------------
int main( int argc, char *argv[] ) {
   char sopt[] = "v";
   int c;
   while( -1 != ( c = getopt( argc, argv, sopt ) ) )
      switch( c ) {
         case 'v':
            debug_level++;
            break;
         default :
            cout << "допустимы только опции: " << sopt << endl;
            return 1;
      }
   if( argc - optind > 2 ) {
      cout << "число параметров программы (" 
           << to_string( argc - 1 ) << ")" << endl;
      return 2;
   }
   ifstream fin;
   progz z;
   if( argc > 1 ) { 
      fin.open( argv[ optind ] );          // на случай -v
      if( !fin ) {
         cout << "файл программы не найден" << endl;      
         return 3;
      }
      fin >> z;
      fin.close();
   }
   while( true ) {                         // командный цикл
      string e;
      cout << '.';
      getline( cin, e );
      if( !e.length() ) e = "h";
      if( !cin ) e = "e";
      const string mode( "tivgsbrl?eqh" );
      int icmd = mode.find( tolower( e[ 0 ] ) ); 
      e = strtrim( e.substr( 1 ) );
      switch( icmd >= 0 ? mode[ icmd ] : '?' ) {
         case 't' :                        // исходный код
            cout << z;
            break;
         case 'i' :                        // отобразить внутреннее представление
            z.put_opers();
            break;
         case 'v' :                        // показать таблицу переменных
            z.put_vars();
            break;
         case 'g' :                        // показать таблицу меток операторов
            z.put_lbls();
            break;
         case 's' :                        // выполнить шаг программы
            z.do_step();
            break;
         case 'b' :                        // выполнить шаг программы
            z.begin();
            break;
         case 'r' :                        // выполнить программу
            try { z.exec( stoi( e ) ); } 
            catch(...) { 
               if( e.length() )
                  cout << z.run_msg( "ошибочный параметр" ) << endl; 
               z.exec();
            }
            break;
         case 'l' :                        // загрузить файл новой программы
            if( !e.length() )
               cout << z.run_msg( "не указан файл программы" ) << endl; 
            else {
               fin.open( e.c_str() );
               if( !fin ) 
                  cout << z.run_msg( "не найден файл программы: " + e ) << endl;
               else
                  fin >> z;
               fin.close();
            }
            break;
         case '?' :                        // версия программы
#define VERSION "0.28"
            cout << "версия " << VERSION << endl;
            break;
         case 'e' :                        // завершение интерпретатора 
         case 'q' :
            z.clean();
            return 0;
         default :
            cout << z.run_msg( "недопустимая команда" ) << endl; 
         case 'h' :
            cout << "допустимые команды:";
            for( char c : mode ) cout << ' ' << c;
            cout << endl;
      }
   }
}